#include <MultiThreading/ThreadPool.h>
#include <iostream>

using namespace MultiThreading;

ThreadPool::ThreadPool(unsigned int thread_count) {
    for (unsigned int i = 0; i < thread_count; i++)
        threads.push_back(new Thread());
}

ThreadPool::ThreadPool() {
    for (unsigned int i = 0; i < std::thread::hardware_concurrency(); i++)
        threads.push_back(new Thread());
}

void ThreadPool::Enqueue(const std::shared_ptr<MultiThreading::TaskBase>& task) {
    std::lock_guard<std::mutex> lock(queue_mutex);

    // Assign it immediately if there's no wait and a thread open.
    std::cout << "qsize (A) " << queue.size() << std::endl;
    if (queue.empty()) {
        for (auto *t: threads) {

            if (t->Busy())
                continue;

            if (!t->SetTask( [this, task] (){ Runner(task); } ))
                throw std::runtime_error("There was an error while setting up the task to run on the thread.");

        }
        return;
    }
    std::cout << "qsize (B) " << queue.size() << std::endl;

    // Alternatively it goes in-to the queue.
    queue.push(task);
}

std::shared_ptr<TaskBase> ThreadPool::Dequeue() {
    std::lock_guard<std::mutex> lock(queue_mutex);
    if (queue.empty())
        return nullptr;

    auto task = queue.front();
    queue.pop();
    return task;
}

void ThreadPool::Enqueue(const std::function<void()>& task) {
    Enqueue(Task<void>::Create(task));
}

void ThreadPool::Runner(const std::shared_ptr<TaskBase>& task) {
    if (!task->Complete())
        task->Run();

    auto next_task = Dequeue();
    if (!next_task)
        return;
    Runner(next_task);
}

unsigned int ThreadPool::QueueSize() {
    std::lock_guard<std::mutex> lock(queue_mutex);
    return queue.size();
}

ThreadPool::~ThreadPool() {
    std::cout << "ThreadPool Destruction Started" << std::endl;


    // Wait for all tasks to be running.
    while (QueueSize() != 0) {}

    std::cout << "All Tasks Complete" << std::endl;

    // delete t waits for the thread to exit gracefully.
    for (auto* t: threads)
        delete t;

    std::cout << "Threads deallocated" << std::endl;


    std::cout << "ThreadPool Destruction Complete" << std::endl;
}
